#
# A minimal example of setup.py to install a Python module
# together with the custom C extension module generated by CFFI.
#

from distutils.core import setup
from distutils.extension import Extension
import bsdopendirtype

setup(name='bsdopendirtype',
      py_modules=['bsdopendirtype'],
      ext_modules=[bsdopendirtype.ffi.verifier.get_extension()])
